/*
 *	EVAL.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */

#include "state.h"

#define DOUBLED_PAWN_PENALTY		(10)
#define ISOLATED_PAWN_PENALTY		(20)
#define BACKWARDS_PAWN_PENALTY	    (8)
#define PASSED_PAWN_BONUS			(20)
#define ROOK_SEMI_OPEN_FILE_BONUS	(10)
#define ROOK_OPEN_FILE_BONUS		(15)
#define ROOK_ON_SEVENTH_BONUS		(20)


/* the values of the pieces */
int piece_value[6] = {
	100, 300, 300, 500, 900, 0
};

/* The "pcsq" arrays are piece/square tables. They're values
   added to the material value of the piece based on the
   location of the piece. */

int pawn_pcsq[64] = {
	  0,   0,   0,   0,   0,   0,   0,   0,
	  5,  10,  15,  20,  20,  15,  10,   5,
	  4,   8,  12,  16,  16,  12,   8,   4,
	  3,   6,   9,  12,  12,   9,   6,   3,
	  2,   4,   6,   8,   8,   6,   4,   2,
	  1,   2,   3, -10, -10,   3,   2,   1,
	  0,   0,   0, -40, -40,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,   0
};

int knight_pcsq[64] = {
	-10, -10, -10, -10, -10, -10, -10, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   0,   5,   5,   5,   5,   0, -10,
	-10,   0,   5,  10,  10,   5,   0, -10,
	-10,   0,   5,  10,  10,   5,   0, -10,
	-10,   0,   5,   5,   5,   5,   0, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10, -30, -10, -10, -10, -10, -30, -10
};

int bishop_pcsq[64] = {
	-10, -10, -10, -10, -10, -10, -10, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   0,   5,   5,   5,   5,   0, -10,
	-10,   0,   5,  10,  10,   5,   0, -10,
	-10,   0,   5,  10,  10,   5,   0, -10,
	-10,   0,   5,   5,   5,   5,   0, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10, -10, -20, -10, -10, -20, -10, -10
};

int king_pcsq[64] = {
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-20, -20, -20, -20, -20, -20, -20, -20,
	  0,  20,  40, -20,   0, -20,  40,  20
};

int king_endgame_pcsq[64] = {
	  0,  10,  20,  30,  30,  20,  10,   0,
	 10,  20,  30,  40,  40,  30,  20,  10,
	 20,  30,  40,  50,  50,  40,  30,  20,
	 30,  40,  50,  60,  60,  50,  40,  30,
	 30,  40,  50,  60,  60,  50,  40,  30,
	 20,  30,  40,  50,  50,  40,  30,  20,
	 10,  20,  30,  40,  40,  30,  20,  10,
	  0,  10,  20,  30,  30,  20,  10,   0
};

/* The flip array is used to calculate the piece/square
   values for DARK pieces. The piece/square value of a
   LIGHT pawn is pawn_pcsq[sq] and the value of a DARK
   pawn is pawn_pcsq[flip[sq]] */
int flip[64] = {
	 56,  57,  58,  59,  60,  61,  62,  63,
	 48,  49,  50,  51,  52,  53,  54,  55,
	 40,  41,  42,  43,  44,  45,  46,  47,
	 32,  33,  34,  35,  36,  37,  38,  39,
	 24,  25,  26,  27,  28,  29,  30,  31,
	 16,  17,  18,  19,  20,  21,  22,  23,
	  8,   9,  10,  11,  12,  13,  14,  15,
	  0,   1,   2,   3,   4,   5,   6,   7
};

int mg_value[6] = { 82, 337, 365, 477, 1025, 0 };
int eg_value[6] = { 94, 281, 297, 512, 936, 0 };

// piece/square tables
// values from Rofchade: http://www.talkchess.com/forum3/viewtopic.php?f=2&t=68311&start=10

int mg_pawn_table[64] = {
	  0,   0,   0,   0,   0,   0,  0,   0,
	 98, 134,  61,  95,  68, 126, 34, -11,
	 -6,   7,  26,  31,  65,  56, 25, -20,
	-14,  13,   6,  21,  23,  12, 17, -23,
	-27,  -2,  -5,  12,  17,   6, 10, -25,
	-26,  -4,  -4, -10,   3,   3, 33, -12,
	-35,  -1, -20, -23, -15,  24, 38, -22,
	  0,   0,   0,   0,   0,   0,  0,   0,
};

int eg_pawn_table[64] = {
	  0,   0,   0,   0,   0,   0,   0,   0,
	178, 173, 158, 134, 147, 132, 165, 187,
	 94, 100,  85,  67,  56,  53,  82,  84,
	 32,  24,  13,   5,  -2,   4,  17,  17,
	 13,   9,  -3,  -7,  -7,  -8,   3,  -1,
	  4,   7,  -6,   1,   0,  -5,  -1,  -8,
	 13,   8,   8,  10,  13,   0,   2,  -7,
	  0,   0,   0,   0,   0,   0,   0,   0,
};

int mg_knight_table[64] = {
	-167, -89, -34, -49,  61, -97, -15, -107,
	 -73, -41,  72,  36,  23,  62,   7,  -17,
	 -47,  60,  37,  65,  84, 129,  73,   44,
	  -9,  17,  19,  53,  37,  69,  18,   22,
	 -13,   4,  16,  13,  28,  19,  21,   -8,
	 -23,  -9,  12,  10,  19,  17,  25,  -16,
	 -29, -53, -12,  -3,  -1,  18, -14,  -19,
	-105, -21, -58, -33, -17, -28, -19,  -23,
};

int eg_knight_table[64] = {
	-58, -38, -13, -28, -31, -27, -63, -99,
	-25,  -8, -25,  -2,  -9, -25, -24, -52,
	-24, -20,  10,   9,  -1,  -9, -19, -41,
	-17,   3,  22,  22,  22,  11,   8, -18,
	-18,  -6,  16,  25,  16,  17,   4, -18,
	-23,  -3,  -1,  15,  10,  -3, -20, -22,
	-42, -20, -10,  -5,  -2, -20, -23, -44,
	-29, -51, -23, -15, -22, -18, -50, -64,
};

int mg_bishop_table[64] = {
	-29,   4, -82, -37, -25, -42,   7,  -8,
	-26,  16, -18, -13,  30,  59,  18, -47,
	-16,  37,  43,  40,  35,  50,  37,  -2,
	 -4,   5,  19,  50,  37,  37,   7,  -2,
	 -6,  13,  13,  26,  34,  12,  10,   4,
	  0,  15,  15,  15,  14,  27,  18,  10,
	  4,  15,  16,   0,   7,  21,  33,   1,
	-33,  -3, -14, -21, -13, -12, -39, -21,
};

int eg_bishop_table[64] = {
	-14, -21, -11,  -8, -7,  -9, -17, -24,
	 -8,  -4,   7, -12, -3, -13,  -4, -14,
	  2,  -8,   0,  -1, -2,   6,   0,   4,
	 -3,   9,  12,   9, 14,  10,   3,   2,
	 -6,   3,  13,  19,  7,  10,  -3,  -9,
	-12,  -3,   8,  10, 13,   3,  -7, -15,
	-14, -18,  -7,  -1,  4,  -9, -15, -27,
	-23,  -9, -23,  -5, -9, -16,  -5, -17,
};

int mg_rook_table[64] = {
	 32,  42,  32,  51, 63,  9,  31,  43,
	 27,  32,  58,  62, 80, 67,  26,  44,
	 -5,  19,  26,  36, 17, 45,  61,  16,
	-24, -11,   7,  26, 24, 35,  -8, -20,
	-36, -26, -12,  -1,  9, -7,   6, -23,
	-45, -25, -16, -17,  3,  0,  -5, -33,
	-44, -16, -20,  -9, -1, 11,  -6, -71,
	-19, -13,   1,  17, 16,  7, -37, -26,
};

int eg_rook_table[64] = {
	13, 10, 18, 15, 12,  12,   8,   5,
	11, 13, 13, 11, -3,   3,   8,   3,
	 7,  7,  7,  5,  4,  -3,  -5,  -3,
	 4,  3, 13,  1,  2,   1,  -1,   2,
	 3,  5,  8,  4, -5,  -6,  -8, -11,
	-4,  0, -5, -1, -7, -12,  -8, -16,
	-6, -6,  0,  2, -9,  -9, -11,  -3,
	-9,  2,  3, -1, -5, -13,   4, -20,
};

int mg_queen_table[64] = {
	-28,   0,  29,  12,  59,  44,  43,  45,
	-24, -39,  -5,   1, -16,  57,  28,  54,
	-13, -17,   7,   8,  29,  56,  47,  57,
	-27, -27, -16, -16,  -1,  17,  -2,   1,
	 -9, -26,  -9, -10,  -2,  -4,   3,  -3,
	-14,   2, -11,  -2,  -5,   2,  14,   5,
	-35,  -8,  11,   2,   8,  15,  -3,   1,
	 -1, -18,  -9,  10, -15, -25, -31, -50,
};

int eg_queen_table[64] = {
	 -9,  22,  22,  27,  27,  19,  10,  20,
	-17,  20,  32,  41,  58,  25,  30,   0,
	-20,   6,   9,  49,  47,  35,  19,   9,
	  3,  22,  24,  45,  57,  40,  57,  36,
	-18,  28,  19,  47,  31,  34,  39,  23,
	-16, -27,  15,   6,   9,  17,  10,   5,
	-22, -23, -30, -16, -16, -23, -36, -32,
	-33, -28, -22, -43,  -5, -32, -20, -41,
};

int mg_king_table[64] = {
	-65,  23,  16, -15, -56, -34,   2,  13,
	 29,  -1, -20,  -7,  -8,  -4, -38, -29,
	 -9,  24,   2, -16, -20,   6,  22, -22,
	-17, -20, -12, -27, -30, -25, -14, -36,
	-49,  -1, -27, -39, -46, -44, -33, -51,
	-14, -14, -22, -46, -44, -30, -15, -27,
	  1,   7,  -8, -64, -43, -16,   9,   8,
	-15,  36,  12, -54,   8, -28,  24,  14,
};

int eg_king_table[64] = {
	-74, -35, -18, -18, -11,  15,   4, -17,
	-12,  17,  14,  17,  17,  38,  23,  11,
	 10,  17,  23,  15,  20,  45,  44,  13,
	 -8,  22,  24,  27,  26,  33,  26,   3,
	-18,  -4,  21,  24,  27,  23,   9, -11,
	-19,  -3,  11,  21,  23,  16,   7,  -9,
	-27, -11,   4,  13,  14,   4,  -5, -17,
	-53, -34, -21, -11, -28, -14, -24, -43
};

int mg_table[2][6][64];
int eg_table[2][6][64];

void init_tables() {
	for (int square = 0; square < 64; square++) {
		mg_table[WHITE][PAWN][square] = mg_value[PAWN] + mg_pawn_table[square];
		mg_table[BLACK][PAWN][square] = mg_value[PAWN] + mg_pawn_table[flip[square]];
		mg_table[WHITE][KNIGHT][square] = mg_value[KNIGHT] + mg_knight_table[square];
		mg_table[BLACK][KNIGHT][square] = mg_value[KNIGHT] + mg_knight_table[flip[square]];
		mg_table[WHITE][BISHOP][square] = mg_value[BISHOP] + mg_bishop_table[square];
		mg_table[BLACK][BISHOP][square] = mg_value[BISHOP] + mg_bishop_table[flip[square]];
		mg_table[WHITE][ROOK][square] = mg_value[ROOK] + mg_rook_table[square];
		mg_table[BLACK][ROOK][square] = mg_value[ROOK] + mg_rook_table[flip[square]];
		mg_table[WHITE][QUEEN][square] = mg_value[QUEEN] + mg_queen_table[square];
		mg_table[BLACK][QUEEN][square] = mg_value[QUEEN] + mg_queen_table[flip[square]];
		mg_table[WHITE][KING][square] = mg_king_table[square];
		mg_table[BLACK][KING][square] = mg_king_table[flip[square]];

		eg_table[WHITE][PAWN][square] = eg_value[PAWN] + eg_pawn_table[square];
		eg_table[BLACK][PAWN][square] = eg_value[PAWN] + eg_pawn_table[flip[square]];
		eg_table[WHITE][KNIGHT][square] = eg_value[KNIGHT] + eg_knight_table[square];
		eg_table[BLACK][KNIGHT][square] = eg_value[KNIGHT] + eg_knight_table[flip[square]];
		eg_table[WHITE][BISHOP][square] = eg_value[BISHOP] + eg_bishop_table[square];
		eg_table[BLACK][BISHOP][square] = eg_value[BISHOP] + eg_bishop_table[flip[square]];
		eg_table[WHITE][ROOK][square] = eg_value[ROOK] + eg_rook_table[square];
		eg_table[BLACK][ROOK][square] = eg_value[ROOK] + eg_rook_table[flip[square]];
		eg_table[WHITE][QUEEN][square] = eg_value[QUEEN] + eg_queen_table[square];
		eg_table[BLACK][QUEEN][square] = eg_value[QUEEN] + eg_queen_table[flip[square]];
		eg_table[WHITE][KING][square] = eg_king_table[square];
		eg_table[BLACK][KING][square] = eg_king_table[flip[square]];
	}
}

/* pawn_rank[x][y] is the rank of the least advanced pawn of color x on file
   y - 1. There are "buffer files" on the left and right to avoid special-case
   logic later. If there's no pawn on a rank, we pretend the pawn is
   impossibly far advanced (0 for LIGHT and 7 for DARK). This makes it easy to
   test for pawns on a rank and it simplifies some pawn evaluation code. */
int pawn_rank[2][10];

int piece_mat[2];  /* the value of a side's pieces */
int pawn_mat[2];  /* the value of a side's pawns */

int State::eval()
{
	int square;

	int mg[2];
	int eg[2];
	int gamePhase = 0;

	mg[WHITE] = 0;
	mg[BLACK] = 0;
	eg[WHITE] = 0;
	eg[BLACK] = 0;

	// evaluate each piece

	for (square = 0; square < 64; ++square) {
		if (color[square] == EMPTY)
			continue;
		if (color[square] == WHITE) {
			switch (piece[square]) {
			case PAWN:
				mg[WHITE] += mg_table[WHITE][PAWN][square];
				eg[WHITE] += eg_table[WHITE][PAWN][square];
				break;
			case KNIGHT:
				mg[WHITE] += mg_table[WHITE][KNIGHT][square];
				eg[WHITE] += eg_table[WHITE][KNIGHT][square];
				gamePhase += 1;
				break;
			case BISHOP:
				mg[WHITE] += mg_table[WHITE][BISHOP][square];
				eg[WHITE] += eg_table[WHITE][BISHOP][square];
				gamePhase += 1;
				break;
			case ROOK:
				mg[WHITE] += mg_table[WHITE][ROOK][square];
				eg[WHITE] += eg_table[WHITE][ROOK][square];
				gamePhase += 2;
				break;
			case QUEEN:
				mg[WHITE] += mg_table[WHITE][QUEEN][square];
				eg[WHITE] += eg_table[WHITE][QUEEN][square];
				gamePhase += 4;
				break;
			case KING:
				mg[WHITE] += mg_table[WHITE][KING][square];
				eg[WHITE] += eg_table[WHITE][KING][square];
				break;
			}
		}
		else {
			switch (piece[square]) {
			case PAWN:
				mg[BLACK] += mg_table[BLACK][PAWN][square];
				eg[BLACK] += eg_table[BLACK][PAWN][square];
				break;
			case KNIGHT:
				mg[BLACK] += mg_table[BLACK][KNIGHT][square];
				eg[BLACK] += eg_table[BLACK][KNIGHT][square];
				gamePhase += 1;
				break;
			case BISHOP:
				mg[BLACK] += mg_table[BLACK][BISHOP][square];
				eg[BLACK] += eg_table[BLACK][BISHOP][square];
				gamePhase += 1;
				break;
			case ROOK:
				mg[BLACK] += mg_table[BLACK][ROOK][square];
				eg[BLACK] += eg_table[BLACK][ROOK][square];
				gamePhase += 2;
				break;
			case QUEEN:
				mg[BLACK] += mg_table[BLACK][QUEEN][square];
				eg[BLACK] += eg_table[BLACK][QUEEN][square];
				gamePhase += 4;
				break;
			case KING:
				mg[BLACK] += mg_table[BLACK][KING][square];
				eg[BLACK] += eg_table[BLACK][KING][square];
				break;
			}
		}
	}

	// scale score according to game phase

	int mgScore = mg[WHITE] - mg[BLACK];
	int egScore = eg[WHITE] - eg[BLACK];
	int mgPhase = gamePhase;
	if (mgPhase > 24) mgPhase = 24; // in case of early promotion
	int egPhase = 24 - mgPhase;

	int score = (mgScore * mgPhase + egScore * egPhase) / 24;

	// return score relative to the side to move

	if (side == WHITE)
		return score;
	else
		return -score;
}