// mcts.cpp

#include "node.h"

using namespace std::chrono;

umove searchMC(State *rootstate, int niterations, milliseconds timelimit)
{
    umove mv;
    mv.u = 0;
    Node* rootnode = new Node(mv, nullptr, rootstate);
    umove moveTable[256];
    int scoreTable[256];
    double meanTable[256];
    int moveCount = 0;
    int tempScore;
    umove tempMove;
    double tempMean;

    int iterations = 0;

    system_clock::time_point old = system_clock::now();

    while (duration_cast<milliseconds>(system_clock::now() - old) < timelimit || iterations < niterations)
    {
        iterations++;

        Node* node = rootnode;
        State* state = new State(*rootstate);

        // Select
        if (node->untriedMoves.empty() && !node->childNodes.empty())
        {
            node = node->selectChild();
            state->makemove(node->mv.b);
        }

        // Expand
        if (!node->untriedMoves.empty())
        {
            umove m = node->untriedMoves[0];
            state->makemove(m.b);
            node = node->addChild(m,state);
        }

        // Simulation or Scoring
        double g = 1.0 - state->simulation();

        // backpropagate
        while (node != nullptr)
        {
            node->update(g);
            g = 1.0 - g;
            node = node->parentNode;
        }

        // periodic display

        if (iterations % 1000 == 0) {

            bool flag = true;
            double score;
            double mean;
            umove oldMove, bestMove;

            for (auto& n : rootnode->childNodes)
            {
                umove m = n->mv;
                double sc = n->visits;

                if (flag)
                {
                    score = sc;
                    oldMove = bestMove;
                    bestMove = m;
                    mean = n->mean;
                    flag = false;
                }

                if (sc > score)
                {
                    score = sc;
                    bestMove = m;
                }
            }

            if (iterations % 10000 == 0 || oldMove.u != bestMove.u)

            printf("%d %d %d %d ",
                1, int(mean), 0, iterations);
            std::cout << rootstate->move_str(bestMove.b) << std::endl;
        }

        delete state;
    }

    std::cout << "iterations: " << iterations << std::endl;

    // pick best move

    double score = 0;
    umove bestMove;

    bool flag = true;

    for (auto& n : rootnode->childNodes)
    {
        umove m = n->mv;
        double sc = n->visits;
        double mean = n->mean;

        moveTable[moveCount] = m;
        scoreTable[moveCount] = sc;
        meanTable[moveCount] = mean;
        moveCount++;

        if (flag)
        {
            score = sc;
            bestMove = m;
            flag = false;
        }

        if (sc > score)
        {
            score = sc;
            bestMove = m;
        }
    }

    // sort moves for display

    for (int ble = 0; ble < moveCount; ble++) {
        for (int i = 0; i < moveCount-1; i++)
        {
            if (scoreTable[i] > scoreTable[i+1]) {
                tempScore = scoreTable[i];
                tempMove = moveTable[i];
                tempMean = meanTable[i];

                scoreTable[i] = scoreTable[i+1];
                moveTable[i] = moveTable[i+1];
                meanTable[i] = meanTable[i+1];
                
                scoreTable[i+1] = tempScore;
                moveTable[i+1] = tempMove;
                meanTable[i + 1] = tempMean;
            }
        }
    }

    // print moves

    for (int i = 0; i < moveCount; i++)
    {
        umove m = moveTable[i];
        double sc = scoreTable[i];        
        std::cout << move2str(m.b) << " " << scoreTable[i] << " visits, mean score " << meanTable[i] << std::endl;
    }

    delete rootnode;

    return bestMove;
}
