#include "node.h"
#include <math.h>

Node::Node(umove m, Node* parent, State* state)
    : mv(m)
    , visits(0)
    , score(0)
    , mean(0)
    , parentNode(parent)
{
    state->genLegal(untriedMoves); // see utility.cpp
}

Node *Node::selectChild()
{
    assert(childNodes.size() > 0);

    Node* selected = nullptr;

    double bestValue = (double)INT_MIN;

    for (auto& n : childNodes)
    {
        assert(visits > 0);

        double utc = n->mean
                   + UCB_CONST * sqrt(log(visits)/n->visits);

        if (utc > bestValue)
        {
            selected = n;
            bestValue = utc;
        }
    }

    assert(selected != nullptr);

    return selected;
}

Node *Node::addChild(umove m, State *state)
{
    Node* n = new Node(m, this, state);

     for (auto it = untriedMoves.begin(); it != untriedMoves.end(); )
     {
         umove um = *it;

         if (um.u == m.u)
         {
             untriedMoves.erase(it);
             break;
         } 
		 else 
		 {
             ++it;
         }
     }

     childNodes.push_back(n);

     return n;
}

void Node::update(double result)
{
    score += result;
    visits++;
    mean = score/visits;
}

Node::~Node()
{
    for (auto &n : childNodes)
    {
        delete n;
    }
}
