/*
 *	MAIN.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */

#include "state.h"

/* get_ms() returns the milliseconds elapsed since midnight,
   January 1, 1970. */

#include <sys/timeb.h>

BOOL ftime_ok = FALSE;  /* does ftime return milliseconds? */

int get_ms()
{
	struct timeb timebuffer;
	ftime(&timebuffer);
	if (timebuffer.millitm != 0)
		ftime_ok = TRUE;
	return (timebuffer.time * 1000) + timebuffer.millitm;
}

static void sortMove(U64 k, std::vector<gen_t>& list)
{
    umove temp;
    U64 index = 0;
    int bestScore = INT_MIN;
    U64 bestNum = k;

    for (index = k; index < list.size(); index++)
    {
        if (list.at(index).score > bestScore)
        {
            bestScore = list.at(index).score;
            bestNum = index;
        }
    }

    temp = list[k].m;
    list[k].m = list[k].m;
    list[k].m = temp;
}


int State::genLegal(std::vector<umove>& list)
{
    list.clear();

    std::vector<gen_t> temp;
    temp.clear();

    ply = 0;
    gen();

    for (int i = ply; i < first_move[ply+1]; i++)
    {
        move_bytes mv = gen_dat[i].m.b;

        if (!makemove(mv))
            continue;
        gen_dat[i].score = eval();
        takeback();

       // if (gen_dat[i].score > alpha && gen_dat[i].score < beta)
        {
            temp.push_back(gen_dat[i]);
        }
    }

    for (U64 i = 0; i < temp.size(); i++)
    {
        sortMove(i, temp);
        list.push_back(temp[i].m);
    }

    return static_cast<int>(list.size());
}

move_bytes State::sudoMove(int f, int t)
{
    for (int i = 0; i < first_move[1]; i++)
    {
        if (gen_dat[i].m.b.from == f && gen_dat[i].m.b.to == t)
        {
            /* if the move is a promotion, handle the promotion piece;
               assume that the promotion moves occur consecutively in
               gen_dat. */
//            if (gen_dat[i].m.b.bits & 32)
//            {
//                return i + 3; // queen promotion only
//            }
            return gen_dat[i].m.b;
        }
    }

    /* didn't find the move */
    return gen_dat[first_move[1]].m.b;
}

bool State::sudoMove(int f, int t, move_bytes& mb)
{
    for (int i = ply; i < first_move[ply+1]; i++)
    {
        if (gen_dat[i].m.b.from == f && gen_dat[i].m.b.to == t)
        {
            mb = gen_dat[i].m.b;
            return true;
        }
    }
    /* didn't find the move */
    return false;
}

/* parse the move s (in coordinate notation) and return the move's
   index in gen_dat, or -1 if the move is illegal */

int State::parse_move(char *s)
{
	int from, to, i;

	/* make sure the string looks like a move */
	if (s[0] < 'a' || s[0] > 'h' ||
			s[1] < '0' || s[1] > '9' ||
			s[2] < 'a' || s[2] > 'h' ||
			s[3] < '0' || s[3] > '9')
		return -1;

	from = s[0] - 'a';
	from += 8 * (8 - (s[1] - '0'));
	to = s[2] - 'a';
	to += 8 * (8 - (s[3] - '0'));

	for (i = 0; i < first_move[1]; ++i)
		if (gen_dat[i].m.b.from == from && gen_dat[i].m.b.to == to) {

			/* if the move is a promotion, handle the promotion piece;
			   assume that the promotion moves occur consecutively in
			   gen_dat. */
			if (gen_dat[i].m.b.bits & 32)
				switch (s[4]) {
					case 'N':
						return i;
					case 'B':
						return i + 1;
					case 'R':
						return i + 2;
					default:  /* assume it's a queen */
						return i + 3;
				}
			return i;
		}

	/* didn't find the move */
	return -1;
}

std::string translate(int sq)
{
    int c = sq % 8;
    int r = sq / 8;

    std::string s = "";
    s += char(c + 97); // ascii 97 = a
    s += char(7 - r + 49); // ascii 49 = 1

    return s;
}

std::string move2str(move_bytes mb)
{
    extern std::string SN[64];
    return SN[(int)mb.from] + SN[(int)mb.to];
}


/* move_str returns a string with move m in coordinate notation */

char* State::move_str(move_bytes m)
{
	static char str[6];

	char c;

	if (m.bits & 32) {
		switch (m.promote) {
			case KNIGHT:
				c = 'n';
				break;
			case BISHOP:
				c = 'b';
				break;
			case ROOK:
				c = 'r';
				break;
			default:
				c = 'q';
				break;
		}
		sprintf(str, "%c%d%c%d%c",
				COL(m.from) + 'a',
				8 - ROW(m.from),
				COL(m.to) + 'a',
				8 - ROW(m.to),
				c);
	}
	else
		sprintf(str, "%c%d%c%d",
				COL(m.from) + 'a',
				8 - ROW(m.from),
				COL(m.to) + 'a',
				8 - ROW(m.to));
	return str;
}


/* print_board() prints the board */

void State::print_board()
{
	int i;
	
	printf("\n8 ");
	for (i = 0; i < 64; ++i) {
		switch (color[i]) {
			case EMPTY:
				printf(" .");
				break;
			case WHITE:
				printf(" %c", piece_char[piece[i]]);
				break;
			case BLACK:
				printf(" %c", piece_char[piece[i]] + ('a' - 'A'));
				break;
		}
		if ((i + 1) % 8 == 0 && i != 63)
			printf("\n%d ", 7 - ROW(i));
	}
	printf("\n\n   a b c d e f g h\n\n");
    fflush(stdout);
}

/* reps() returns the number of times the current position
   has been repeated. It compares the current value of hash
   to previous values. */

int State::reps()
{
    int r = 0;

    for (int i = hply - fifty; i < hply; i++)
    {
        if (hist_dat[i].hash == hash)
        {
            r++;
        }
    }

    return r;
}

bool State::isGameOver()
{
    int i = 0;

    /* is there a legal move? */
    for (i = 0; i < first_move[1]; i++)
    {
        if (makemove(gen_dat[i].m.b)) {
            takeback();
            break;
        }
    }

    if (i == first_move[1])
    {
        if (in_check(side))
        {
            if (side == WHITE)
            {
                printf("0-1 {Black mates}\n"); fflush(stdout);
                return true;
            }
            else
            {
                printf("1-0 {White mates}\n");
                return true; fflush(stdout);
            }
        }
        else
        {
            printf("1/2-1/2 {Stalemate}\n");
            return true; fflush(stdout);
        }
    }
    else if (reps() == 3)
    {
        printf("1/2-1/2 {Draw by repetition}\n"); fflush(stdout);
        return true;
    }
    else if (fifty >= 100)
    {
        printf("1/2-1/2 {Draw by fifty move rule}\n"); fflush(stdout);
        return true;
    }
    return false;
}

void State::printStatus()
{
    std::cout << "Status..." << std::endl;
    std::cout << "side to move -> " << (side == WHITE? "LIGHT":"DARK") << std::endl;
    std::cout << "ep -> " << ep << std::endl;
    std::cout << "castle -> " << castle << std::endl;
    std::cout << "fifty -> " << fifty << std::endl;
    std::cout << "hash -> " << hash << std::endl;
    std::cout << "ply -> " << ply << std::endl;
    std::cout << "hply -> " << hply << std::endl;
    std::cout << std::endl;
}

double sigmoid(int score) {
    double exp = -(1.00 * ((double)score) / 400.0);
    return 1.0 / (1.0 + pow(10.0, exp));
}
